/*
 pattern files can be retrieved from: http://entropymine.com/jason/life/p/jslife-20051021.zip
 place the files in /sys/games/lib/life (without the subdirectories).
 pattern files from other sources are likely in a slightly different format (or same format but different semantics...).
 by mechiel lukkien, mechiel@xs4all.nl, january 2006.
*/

#include <u.h>
#include <libc.h>
#include <ctype.h>
#include <bio.h>
#include <draw.h>
#include <event.h>

typedef struct Board Board;

struct Board
{
	char **table;
	int nrows, ncols;
};

char *patterndir = "/sys/games/lib/life";

Image *deadimg, *aliveimg, *bgimage;
int brdi;
Board brds[2];
char *genpatterns(int);
int curpattern;
char **files;
int nfiles;
int speed = 8;
int counter;
int gridsize = 3;

Menu lmenu = {
	nil,
	genpatterns,
	0,
};

char *buttons[] = {
	"start",
	"stop",
	"clear",
	"slower",
	"faster",
	"smaller",
	"larger",
	"write",
	"exit",
	nil,
};

Menu menu = {
	buttons
};


void *
emalloc(int s)
{
	void *p;
	p = malloc(s);
	if(!p)
		sysfatal("malloc");
	return p;
}

void
readpatternlist(void)
{
	int fd;
	Dir *d;
	int i;

	fd = open(patterndir, OREAD);
	if(fd < 0 || (nfiles = dirreadall(fd, &d)) < 0)
		sysfatal("reading files: %r");
	files = emalloc(sizeof files[0] * nfiles);
	for(i = 0; i < nfiles; i++)
		files[i] = d[i].name;
	close(fd);
}

char *
genpatterns(int i)
{
	if(i >= 0 && i < nfiles)
		return files[i];
	return nil;
}

void
reallocboard(Board *b, int nrows, int ncols)
{
	int i;

	for(i = 0; i < b->nrows; i++)
		free(b->table[i]);
	free(b->table);

	b->table = emalloc(nrows * sizeof b->table[0]);
	for(i = 0; i < nrows; i++)
		b->table[i] = emalloc(ncols);
	b->nrows = nrows;
	b->ncols = ncols;
}

void
rowset(Board *b, int row, int col, int n, int status)
{
	while(col < b->ncols && n > 0){
		b->table[row][col++] = status;
		n--;
	}
}

/* parse .rle files.  the collection i used named them .lif, though that's another format.  this format is also not uniformly interpreted by different collection authors! */
void
parse(char *file, Board *b)
{
	Biobuf *bp;
	char *l, *p;
	char *args[11];
	int n;
	int nrows, ncols, havedim;
	int row, col;
	int nr;

	bp = Bopen(file, OREAD);
	if(!bp)
		sysfatal("Bopen: %r");

	l = nil;
	nrows = ncols = row = col = 0;
	havedim = 0;
	nr = 0;
	for(;;){
		free(l);
		l = Brdstr(bp, '\n', 1);
		if(!l)
			sysfatal("unexpected end of file");
		nr++;

		if(l[0] == '#')
			continue;
		if(!havedim){
			n = getfields(l, args, nelem(args), 0, " \t,");
			if(n < 7 || strcmp(args[0], "x") != 0 || strcmp(args[1], "=") != 0)
				sysfatal("%s:%d: invalid dimensions line (x =)", file, nr);
			if(strcmp(args[3], "") != 0 || strcmp(args[4], "y") != 0 || strcmp(args[5], "=") != 0)
				sysfatal("%s:%d: invalid dimensions line (y =)", file, nr);
			if(n > 7 && (strcmp(args[7], "") != 0 || strcmp(args[8], "rule") != 0 || strcmp(args[9], "=") != 0 || strcmp(args[10], "B3/S23") != 0))
				sysfatal("%s:%d: invalid dimensions line (rule = B3/S23)", file, nr);
			ncols = atoi(args[2]);
			nrows = atoi(args[6]);
			havedim = 1;
			reallocboard(brds, nrows, ncols);
			reallocboard(brds+1, nrows, ncols);
		} else {
			p = l;
			while(p[0]){
				if(col == ncols){
					row++;
					col = 0;
				}
				if(p[0] != '!' && row >= nrows)
					sysfatal("%s:%d: too many rows in file", file, nr);
				n = 1;
				if(isdigit(p[0])){
					n = atoi(p);
					p += strspn(p, "0123456789");
				}
				switch(p[0]){
				case 'o':	/* alive cell */
				case 'b':	/* dead cell */
					if(col + n > ncols)
						sysfatal("%s:%d: row not long enough (n=%d type=%c)", file, nr, n, p[0]);
					rowset(b, row, col, n, (p[0] == 'o') ? 1 : 0);
					col += n;
					if(col == ncols && p[1] == '$')
						p++;
					break;
				case '$':
					rowset(b, row, col, ncols - col, 0);
					row++;
					col = 0;
					break;
				case '!':
					while(row < nrows){
						rowset(b, row, col, ncols - col, 0);
						row++;
						col = 0;
					}
					goto stop;
					break;
				default:
					sysfatal("%s:%d: invalid character in file: %d\n", file, nr, p[0]);
				}
				p++;
			}
		}			
	}
stop:
	free(l);

	Bterm(bp);
}

void
dump(Board *b)
{
	int i, j;

	for(i = 0; i < b->nrows; i++) {
		for(j = 0; j < b->ncols; j++)
			print(b->table[i][j] ? "#" : " ");
		print("\n");
	}
}

int
get(Board *b, int row, int col)
{
	if(row < 0 || col < 0 || row >= b->nrows || col >= b->ncols)
		return 0;
	return b->table[row][col];
}

int
health(Board *b, int row, int col)
{
	int alive;
	int i, j;

	alive = 0;
	for(i = -1; i <= 1; i++)
		for(j = -1; j <= 1; j++) {
			if(i == 0 && j == 0)
				continue;
			alive += get(b, row+i, col+j);
		}
	;

	if(!b->table[row][col] && alive == 3)
		return 1;
	if(b->table[row][col] && (alive == 2 || alive == 3))
		return 1;
	return 0;
}

void
run(Board *srcb, Board *dstb)
{
	int i, j;

	for(i = 0; i < srcb->nrows; i++)
		for(j = 0; j < srcb->ncols; j++)
			dstb->table[i][j] = health(srcb, i, j);
}

void
drawlife(int drawall)
{
	Rectangle r;
	int i, j;
	Board *b;

	r = rectaddpt(Rect(0, 0, gridsize, gridsize), screen->r.min);

	b = brds+brdi;
	for(i = 0; i < b->nrows; i++)
		for(j = 0; j < b->ncols; j++)
			if(drawall || brds[brdi].table[i][j] != brds[1-brdi].table[i][j])
				draw(screen, rectaddpt(r, Pt(j*(gridsize+1), i*(gridsize+1))), b->table[i][j] ? aliveimg : deadimg, nil, ZP);
	flushimage(display, 1);
}

void
eresized(int new)
{
	int fd;
	Board *b;

	if(new && getwindow(display, Refnone) < 0)
		sysfatal("getwindow after resize");

	fd = open("/dev/wctl", OWRITE);
	if(fd >= 0){
		b = brds+brdi;
		fprint(fd, "resize -dx %d -dy %d", 8+b->ncols*(gridsize+1), 8+b->nrows*(gridsize+1));
		close(fd);
	}

	draw(screen, screen->r, bgimage, nil, ZP);
	drawlife(1);
}

void
usage(void)
{
	fprint(2, "usage: life [file]\n");
	exits("usage");
}

void
main(int argc, char *argv[])
{
	Event e;
	Mouse m;
	int timer;
	int key;
	char *file;
	int i;
	Board *b;
	int newstatus;
	int fd;

	ARGBEGIN {
	default:
		usage();
	} ARGEND

	if(argc > 1)
		usage();

	readpatternlist();

	if(argc == 0)
		file = smprint("%s/%s", patterndir, files[0]);
	else
		file = strdup(argv[0]);
	parse(file, &brds[brdi]);

	if(initdraw(nil, nil, "life") < 0)
		sysfatal("initdraw failed: %r");

	einit(Emouse|Ekeyboard);
	timer = etimer(0, 100);
	if(timer == 0)
		sysfatal("etimer");

	bgimage = allocimagemix(display, DWhite, DWhite);
	deadimg = allocimagemix(display, DBlack, DWhite);
	aliveimg = allocimagemix(display, DBlack, DBlack);

	eresized(0);

	newstatus = -1;
	for(;;){
		key = event(&e);
		b = brds+brdi;

		if(key == Emouse){
			m = e.mouse;
			if(~m.buttons&1)
				newstatus = -1;
			if(m.buttons&1){
				int row, col;
				Point xy;

				xy = subpt(m.xy, screen->r.min);
				row = xy.y / (gridsize+1);
				col = xy.x / (gridsize+1);
				if(row >= 0 && col >= 0 && row < b->nrows && col < b->ncols){
					if(newstatus == -1)
						newstatus = 1 - b->table[row][col];
					b->table[row][col] = newstatus;
				}
				drawlife(1);
			}
			if(m.buttons&2){
				int l;

				lmenu.lasthit = curpattern;
				l = emenuhit(2, &m, &lmenu);
				if(l >= 0){
					curpattern = l;
					file = smprint("%s/%s", patterndir, files[l]);
					parse(file, b);
					free(file);
				}
				eresized(0);
			}
			if(m.buttons&4){
				switch(emenuhit(3, &m, &menu)){
				case 0:	/* start */
					speed = 8;
					break;
				case 1:	/* stop */
					speed = 0;
					break;
				case 2:	/* clear */
					for(i = 0; i < b->nrows; i++)
						rowset(b, i, 0, b->ncols, 0);
					drawlife(1);
					break;
				case 3:	/* slower */
					speed *= 2;
					break;
				case 4:	/* faster */
					speed /= 2;
					if(speed <= 0)
						speed = 1;
					break;
				case 5:	/* smaller */
					gridsize /= 2;
					if(gridsize <= 0)
						gridsize = 1;
					eresized(0);
					break;
				case 6:	/* larger */
					gridsize *= 2;
					eresized(0);
					break;
				case 7:	/* write */
					fd = create("/tmp/life.image", OWRITE, 0666);
					if(fd > 0){
						if(writeimage(fd, screen, 0) == -1)
							sysfatal("writing image: %r");
						close(fd);
					}
					break;
				case 8:	/* exit */
					exits(nil);
				}
			}
		}

		if(key == Ekeyboard){
			switch(e.kbdc){
			case 127:
			case 'q':
				exits(nil);
			}
		}

		if(key == timer){
			counter++;
			if(speed > 0 && counter % speed == 0){
				run(&brds[brdi], &brds[1 - brdi]);
				brdi = 1 - brdi;
				drawlife(0);
			}
		}
	}
}
